import csv
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler  # z-score归一化表示的包


def get_adjacency_matrix(distance_df_filename, num_of_vertices):
    with open(distance_df_filename, 'r') as f:
        reader = csv.reader(f)
        header = f.__next__()
        edges = [(int(i[0]), int(i[1]), float(i[2])) for i in reader]

    A = np.zeros((int(num_of_vertices), int(num_of_vertices)),
                 dtype=np.float32)

    for i, j, z in edges:
        z = z / 100  # 将权重映射到（0，1）用于下文判别
        z2 = z * z
        A[i, j] = np.exp(-z2 / 10)
        A[j, i] = np.exp(-z2 / 10)

    return A


def weight_matrix(file_path, sigma2=0.1, epsilon=0.5, scaling=True):
    try:
        W = pd.read_csv(file_path, header=None).values
    except FileNotFoundError:
        print(f'ERROR: input file was not found in {file_path}.')

    # check whether W is a 0/1 matrix.
    if set(np.unique(W)) == {0, 1}:
        print('The input graph is a 0/1 matrix; set "scaling" to False.')
        scaling = False

    if scaling:
        n = W.shape[0]
        W = W / 10000.
        W2, W_mask = W * W, np.ones([n, n]) - np.identity(n)
        return np.exp(-W2 / sigma2) * (np.exp(-W2 / sigma2) >= epsilon) * W_mask
    else:
        return W


def get_pemsbay_weekday_data():  # Speed_325 : 2017年3月6日到4月30日 56天，同样需要将周末过滤掉, 对数据实行z-score归一化
    adj_filename = "./data/pemsbay/adj_mx_bay.npy"
    graph_signal_matrix_filename = "./data/pemsbay/Speed_325.csv"
    num_of_vertices = 325
    day_slot = 288
    distance_pattern = np.load(adj_filename)
    data_seq = pd.read_csv(graph_signal_matrix_filename, header=None).values.astype(float)
    data_seq_flow = np.array(data_seq)
    data_scaler = StandardScaler()
    data_seq_flow = data_scaler.fit_transform(data_seq_flow)
    day_sum = int(data_seq_flow.shape[0]/day_slot)
    weekend_index = []
    for i in range(1,9):
        weekend_index.append(7*i-2)
        weekend_index.append(7*i-1)
    data_seq_flow_weekend = np.zeros((day_slot,num_of_vertices))
    data_seq_flow_weekday = np.zeros((day_slot,num_of_vertices))
    countSum = 0  # 用于数据集中休息日的计数，将数据集首先进行分类，进行全局的关系发现
    countSum2 = 0  # 用于数据集中工作日的计数
    for i in range(day_sum):
        data_seq_flow_day = data_seq_flow[day_slot * i: day_slot * (i + 1), :]
        if i in weekend_index:
            data_seq_flow_weekend = data_seq_flow_weekend + np.array(data_seq_flow_day)
            countSum += 1
        else:
            data_seq_flow_weekday = data_seq_flow_weekday + np.array(data_seq_flow_day)
            countSum2 += 1
    weekend_avg = (1 / countSum) * data_seq_flow_weekend
    weekday_avg = (1 / countSum2) * data_seq_flow_weekday
    return distance_pattern, weekday_avg, weekend_avg


def get_pemsm_weekday_data():
    adj_filename = "./data/PeMS-M/W_228.csv"
    graph_signal_matrix_filename = "./data/PeMS-M/V_228.csv"
    num_of_vertices = 228
    day_slot = 288
    distance_pattern = weight_matrix(adj_filename) + np.identity(num_of_vertices)
    data_seq = pd.read_csv(graph_signal_matrix_filename, header=None).values.astype(float)
    data_seq_flow = np.array(data_seq)
    data_scaler = StandardScaler()
    data_seq_flow = data_scaler.fit_transform(data_seq_flow)
    day_sum = int(data_seq_flow.shape[0]/day_slot)
    data_seq_flow_weekday = np.zeros((day_slot,num_of_vertices))
    countSum = 0  # 用于数据集中工作日的计数
    for i in range(day_sum):
        data_seq_flow_day = data_seq_flow[day_slot*i: day_slot*(i+1), :]
        data_seq_flow_weekday = data_seq_flow_weekday + np.array(data_seq_flow_day)
        countSum += 1
    weekday_avg = (1/countSum) * data_seq_flow_weekday
    return distance_pattern, weekday_avg


def get_pems08_weekday_data():
    adj_filename = "./data/PEMSD8/distance.csv"
    graph_signal_matrix_filename = "./data/PEMSD8/Speed_170.csv"
    num_of_vertices = 170
    day_slot = 288
    distance_pattern = get_adjacency_matrix(adj_filename, num_of_vertices) + np.identity(num_of_vertices)
    data_seq = pd.read_csv(graph_signal_matrix_filename, header=None).values.astype(float)
    data_seq_flow = np.array(data_seq)
    data_scaler = StandardScaler()
    data_seq_flow = data_scaler.fit_transform(data_seq_flow)
    day_sum = int(data_seq_flow.shape[0]/day_slot)
    weekend_index = []
    for i in range(1,9):
        weekend_index.append(7*i + 1)
        weekend_index.append(7*i + 2)
    weekend_index.append(1)  # 将第一周的数据分开存放
    weekend_index.append(2)
    data_seq_flow_weekend = np.zeros((day_slot,num_of_vertices))
    data_seq_flow_weekday = np.zeros((day_slot,num_of_vertices))
    countSum = 0  # 用于数据集中休息日的计数
    countSum2 = 0  # 用于数据集中工作日的计数
    for i in range(day_sum):
        data_seq_flow_day = data_seq_flow[day_slot*i: day_slot*(i+1), :]
        if i in weekend_index:
            data_seq_flow_weekend = data_seq_flow_weekend + np.array(data_seq_flow_day)
            countSum += 1
        else:
            data_seq_flow_weekday = data_seq_flow_weekday + np.array(data_seq_flow_day)
            countSum2 += 1
    weekend_avg = (1/countSum) * data_seq_flow_weekend
    weekday_avg = (1/countSum2) * data_seq_flow_weekday
    return distance_pattern, weekday_avg, weekend_avg


pemsm_adj, _ = get_pemsm_weekday_data()
pemsd8_adj, _, _ = get_pems08_weekday_data()
pemsbay_adj, _, _ = get_pemsbay_weekday_data()
adj_text_file_path = ['./data/adj_text_format/PEMSM', './data/adj_text_format/PEMSD8', './data/adj_text_format/PEMSBAY']
df_pemsm_adj = pd.DataFrame(pemsm_adj)
df_pemsd8_adj = pd.DataFrame(pemsd8_adj)
df_pemsbay_adj = pd.DataFrame(pemsbay_adj)
df_pemsm_adj.to_csv(adj_text_file_path[0]+".csv", index=False, header=False)
df_pemsd8_adj.to_csv(adj_text_file_path[1]+".csv", index=False, header=False)
df_pemsbay_adj.to_csv(adj_text_file_path[2]+".csv", index=False, header=False)


# np_pemsd8_adj = np.array(pemsd8_adj)
# n_node = np_pemsd8_adj.shape[0]
# with open(adj_text_file_path[1] + '.txt', 'w') as f:
#     for i in range(n_node):
#         for j in range(n_node):
#             f.write(str(i) + " " + str(j) + " " + str('%.7f' % np_pemsd8_adj[i][j]) + "\n")
# f.close()


# np_pemsbay_adj = np.array(pemsbay_adj)
# n_node = np_pemsbay_adj.shape[0]
# print(n_node)
# with open(adj_text_file_path[2] + '.txt', 'w') as f:
#     for i in range(n_node):
#         for j in range(n_node):
#             f.write(str(i) + " " + str(j) + " " + str('%.7f' % np_pemsbay_adj[i][j]) + "\n")
# f.close()
#
#
# np_pemsm_adj = np.array(pemsm_adj)
# n_node = np_pemsm_adj.shape[0]
# with open(adj_text_file_path[0] + '.txt', 'w') as f:
#     for i in range(n_node):
#         for j in range(n_node):
#             f.write(str(i) + " " + str(j) + " " + str('%.7f' % np_pemsm_adj[i][j]) + "\n")
# f.close()


