import math
import torch
import torch.nn as nn
import torch.nn.init as init

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
torch.set_default_tensor_type(torch.DoubleTensor)
# 使用消融实验证明线图和超图的作用，使用普通图和不使用线图的效果


class HyperGraphConvLayer(nn.Module):
    def __init__(self,input_feature, embedding_size, Lk):
        super(HyperGraphConvLayer, self).__init__()
        self.Lk = Lk
        self.theta = nn.Parameter(torch.DoubleTensor(input_feature, embedding_size))
        self.b = nn.Parameter(torch.DoubleTensor(1,embedding_size, 1, 1))
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        x_c = torch.einsum("nm,bitm->bitn", self.Lk, x)
        x_gc = torch.einsum("io,bitn->botn", self.theta, x_c) + self.b
        return torch.relu(x_gc + x)


class LineGraphConvLayer(nn.Module):
    def __init__(self,input_feature ,embedding_size, n_node, Lk):
        super(LineGraphConvLayer, self).__init__()
        self.Lk = Lk
        self.theta = nn.Parameter(torch.DoubleTensor(input_feature, embedding_size))
        self.Flatten = nn.Linear(self.Lk.shape[0], n_node)
        self.b = nn.Parameter(torch.DoubleTensor(1, embedding_size, 1, 1))
        self.reset_parameters()

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        up_sampling_Lk = self.Flatten(self.Lk)
        x_c = torch.einsum("nm,bitm->bitm", up_sampling_Lk, x)
        x_gc = torch.einsum("io,bitm->botm", self.theta, x_c) + self.b
        return torch.relu(x_gc + x)


class DualChannelGraphConv(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output, n_node, hyper_graph_lap, line_graph_lap, dropout_rate):
        super(DualChannelGraphConv, self).__init__()
        self.hyper_graph_conv_layer1 = HyperGraphConvLayer(input_feature, embedding_size, hyper_graph_lap)
        self.hyper_graph_conv_layer2 = HyperGraphConvLayer(embedding_size, embedding_size, hyper_graph_lap)
        self.line_graph_conv_layer1 = LineGraphConvLayer(input_feature, embedding_size, n_node, line_graph_lap)
        self.line_graph_conv_layer2 = LineGraphConvLayer(embedding_size, embedding_size, n_node, line_graph_lap)
        self.conv_output = nn.Linear(2 * embedding_size, conv_output)

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        x_hyper_embedding_layer_one = self.hyper_graph_conv_layer1(x)  # 交换学习信息,将区域与整体的信息结合起来
        x_line_embedding_layer_one = self.line_graph_conv_layer1(x)
        x_hyper_embedding_layer_output = self.hyper_graph_conv_layer2(x_line_embedding_layer_one)
        x_line_embedding_layer_output = self.line_graph_conv_layer2(x_hyper_embedding_layer_one)
        x_output = torch.cat([x_hyper_embedding_layer_output, x_line_embedding_layer_output], dim=1)
        x_output = x_output.permute(0, 2, 3, 1)
        x_output = self.conv_output(x_output)
        # x_output = x_output.reshape(x_output.shape[0], x_output.shape[1], -1)
        return x_output


class GraphConv(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output, graph_lap):
        super(GraphConv, self).__init__()
        self.graph_conv_layer1 = HyperGraphConvLayer(input_feature, embedding_size, graph_lap)
        self.graph_conv_layer2 = HyperGraphConvLayer(embedding_size, embedding_size, graph_lap)
        self.conv_output = nn.Linear(embedding_size, conv_output)

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        x_1 = self.graph_conv_layer1(x)
        x_output = self.graph_conv_layer2(x_1)
        x_output = x_output.permute(0, 2, 3, 1)
        x_output = self.conv_output(x_output)
        # x_output = x_output.reshape(x_output.shape[0], x_output.shape[1], -1)
        return x_output


class HGraphConv(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output, hyper_graph_lap):
        super(HGraphConv, self).__init__()
        self.h_graph_conv_layer1 = HyperGraphConvLayer(input_feature, embedding_size, hyper_graph_lap)
        self.h_graph_conv_layer2 = HyperGraphConvLayer(embedding_size, embedding_size, hyper_graph_lap)
        self.conv_output = nn.Linear(embedding_size, conv_output)

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        x_1 = self.h_graph_conv_layer1(x)
        x_output = self.h_graph_conv_layer2(x_1)
        x_output = x_output.permute(0, 2, 3, 1)
        x_output = self.conv_output(x_output)
        # x_output = x_output.reshape(x_output.shape[0], x_output.shape[1], -1)
        return x_output


class LGraphConv(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output, n_node, line_graph_lap):
        super(LGraphConv, self).__init__()
        self.l_graph_conv_layer1 = LineGraphConvLayer(input_feature, embedding_size, n_node, line_graph_lap)
        self.l_graph_conv_layer2 = LineGraphConvLayer(embedding_size, embedding_size, n_node, line_graph_lap)
        self.conv_output = nn.Linear(embedding_size, conv_output)

    def reset_parameters(self):
        init.kaiming_uniform_(self.theta, a=math.sqrt(5))
        fan_in, _ = init._calculate_fan_in_and_fan_out(self.theta)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(self.b, -bound, bound)

    def forward(self, x):
        x_1 = self.l_graph_conv_layer1(x)
        x_output = self.l_graph_conv_layer2(x_1)
        x_output = x_output.permute(0, 2, 3, 1)
        x_output = self.conv_output(x_output)
        # x_output = x_output.reshape(x_output.shape[0], x_output.shape[1], -1)
        return x_output


# input X shape, bitm[50，1，12，207]
# b 表示训练的批次大小，i 表示输入矩阵的特征个数， t 表示历史输入的时间序列长度， m 表示研究目标路网中包含多少个路段
# label Y shape, btm[50, 3, 207]
# b 表示训练的批次大小， t 表示将要预测的时间序列长度， m 表示研究目标路网中包含多少个路段
# 模拟输入输出测试模型是否可以调试
# x_fake = torch.Tensor(50, 1, 12, 207).to(device)
# hyper_graph_lap = construct_metrla_hyper_graph_laplacian(64)
# line_graph_lap = construct_metrla_line_graph_from_hyper_graph(64)
# n_node = x_fake.shape[3]
# dual_channel_graph_conv = DualChannelGraphConv(1, 32, n_node, hyper_graph_lap, line_graph_lap).to(device)
# x_predict = dual_channel_graph_conv(x_fake)
# print(x_predict.shape)
