import torch.nn as nn
from dual_hypergraph_convolution import DualChannelGraphConv, GraphConv, LGraphConv, HGraphConv
from seq_transformer import SeqTransformerModel
from seq_transformer_fix_position import SeqTransformerModelFixPos
from seq_transformer_with_parameter import SeqTransformerModelWeight


class HyperGraphTransformer(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, n_node, hyper_graph_lap, line_graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(HyperGraphTransformer, self).__init__()
        self.SBlock = DualChannelGraphConv(input_feature, embedding_size, conv_output_size,
                                           n_node, hyper_graph_lap, line_graph_lap)
        self.TBlock = SeqTransformerModel(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1).unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output).squeeze()


class HyperGraphTransformerWeight(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, n_node, hyper_graph_lap, line_graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(HyperGraphTransformerWeight, self).__init__()
        self.SBlock = DualChannelGraphConv(input_feature, embedding_size, conv_output_size,
                                           n_node, hyper_graph_lap, line_graph_lap)
        self.TBlock = SeqTransformerModelWeight(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output, weight_enc = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output).squeeze(), weight_enc

class GraphTransformer(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(GraphTransformer, self).__init__()
        self.SBlock = GraphConv(input_feature, embedding_size, conv_output_size, graph_lap)
        self.TBlock = SeqTransformerModel(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output)


class HGraphTransformer(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, hyper_graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(HGraphTransformer, self).__init__()
        self.SBlock = HGraphConv(input_feature, embedding_size, conv_output_size, hyper_graph_lap)
        self.TBlock = SeqTransformerModel(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output)


class LGraphTransformer(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, n_node, line_graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(LGraphTransformer, self).__init__()
        self.SBlock = LGraphConv(input_feature, embedding_size, conv_output_size, n_node, line_graph_lap)
        self.TBlock = SeqTransformerModel(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output)


class HyperGraphTransformerFixPos(nn.Module):
    def __init__(self, input_feature, embedding_size, conv_output_size, n_node, hyper_graph_lap, line_graph_lap,
                 hidden_size, filter_size, dropout_rate, n_layers, target_len):  # hidden_size为节点的倍数
        super(HyperGraphTransformerFixPos, self).__init__()
        self.SBlock = DualChannelGraphConv(input_feature, embedding_size, conv_output_size,
                                           n_node, hyper_graph_lap, line_graph_lap)
        self.TBlock = SeqTransformerModelFixPos(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)  # dual channel convolution, 乘2

    def forward(self, x_input, dec_inputs):
        enc_input = self.SBlock(x_input)
        src_dec_input = dec_inputs.unsqueeze(1)
        dec_inputs = self.SBlock(src_dec_input)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output)


class TransformerModel(nn.Module):
    def __init__(self, conv_output_size, hidden_size, filter_size, dropout_rate, n_layers, target_len):
        # hidden_size为节点的倍数
        super(TransformerModel, self).__init__()
        self.TBlock = SeqTransformerModel(hidden_size, filter_size, dropout_rate, n_layers,target_len)
        self.output_layer = nn.Linear(hidden_size, hidden_size//conv_output_size)

    def forward(self, x_input, dec_inputs):
        x_input = x_input.permute(0, 2, 3, 1)
        enc_input = x_input.reshape(x_input.shape[0], x_input.shape[1], -1)
        output = self.TBlock(enc_input, dec_inputs)
        return self.output_layer(output)